import pybullet
import time
import random
import os
import numpy as np
import pybullet_data
import pybullet_planning

RED = [1., 0., 0., 1.]
GREEN = [0., 1., 0., 1.]
BLUE = [0., 0., 1., 1.]
YELLOW = [1., 1., 0., 1.]
TRANSPARENT_PURPLE = [1., 0., 1., 0.5]

SMALL = 0.1
MEDIUM = 0.5
BIG = 1.

DUCK = "duck"
CUBE = "cube"
WEIGHT = "weight"
BEAR = "bear"

object_dict = {
    "duck": ("duck.obj", "duck_vhacd.obj"),
    "cube": ("cube.obj", "cube.obj"),
    "weight": ("stone.obj", "stone.obj"),
    "bear":("teddy2_VHACD_CHs.obj", "teddy2_VHACD_CHs.obj")
}

def initialize_object(object, size=SMALL):
    try:
        object_tuple = object_dict[object]
    except:
        raise ValueError("""You inputted an invalid object. Make sure you specify a case-sensitive object 
                            from the following list: DUCK, CUBE, WEIGHT, BEAR""")

    pybullet.connect(pybullet.GUI)

    pybullet.setAdditionalSearchPath(pybullet_data.getDataPath())
    pybullet.loadURDF("plane_transparent.urdf", useMaximalCoordinates=True)
    #disable rendering during creation.
    pybullet.configureDebugVisualizer(pybullet.COV_ENABLE_GUI, 0)

    shift1 = [0, 0.0, 0]
    shift2 = [0, 0, 0]

    meshScale = [size, size, size]

    # initialize a duck
    visualShapeId = pybullet.createVisualShape(shapeType=pybullet.GEOM_MESH,
                                               fileName=object_tuple[0],
                                               meshScale=meshScale)
    collisionShapeId = pybullet.createCollisionShape(shapeType=pybullet.GEOM_MESH,
                                                     fileName=object_tuple[1],
                                                     meshScale=meshScale)

    assert(visualShapeId == collisionShapeId == 0)

    mb = pybullet.createMultiBody(baseMass=1,
                                  baseInertialFramePosition=[0, 0, 0],
                                  baseCollisionShapeIndex=collisionShapeId,
                                  baseVisualShapeIndex=visualShapeId,
                                  useMaximalCoordinates=False)

    pybullet.changeVisualShape(mb, -1, rgbaColor=[1, 1, 1, 1])
    pybullet.configureDebugVisualizer(pybullet.COV_ENABLE_RENDERING, 1)
    pybullet.setGravity(0, 0, -10)
    pybullet.setRealTimeSimulation(1)
    return mb

def change_to_colors(colors):
    current_color = 0
    while (1):
        keys = pybullet.getKeyboardEvents()
        if keys and len(colors) > 0:
            key = list(keys.keys())[0]
            if keys[key] == pybullet.KEY_WAS_RELEASED:
                current_color = (current_color + 1) % len(colors)
                pybullet.changeVisualShape(1, -1, rgbaColor=colors[current_color])
                # adapted from pybullet demo: create_visual_shape_array.py

def initialize_environment():
    pybullet.connect(pybullet.GUI)
    pybullet.setAdditionalSearchPath(pybullet_data.getDataPath())
    pybullet.loadURDF("plane_transparent.urdf", useMaximalCoordinates=True)
    #disable rendering during creation.
    pybullet.configureDebugVisualizer(pybullet.COV_ENABLE_GUI, 0)
    pybullet.configureDebugVisualizer(pybullet.COV_ENABLE_RENDERING, 1)
    pybullet.setGravity(0, 0, -10)
    pybullet.setRealTimeSimulation(1)

def spawn_object(obj, size, color):
    try:
        object_tuple = object_dict[obj]
    except:
        raise ValueError("""You inputted an invalid object. Make sure you specify a case-sensitive object 
                            from the following list: DUCK, CUBE, WEIGHT, BEAR""")

    meshScale = [size, size, size]

    # initialize a duck
    visual_id = pybullet.createVisualShape(shapeType=pybullet.GEOM_MESH,
                                            fileName=object_tuple[0],
                                            meshScale=meshScale)
    collision_id = pybullet.createCollisionShape(shapeType=pybullet.GEOM_MESH,
                                                fileName=object_tuple[1],
                                                meshScale=meshScale)

    multibody_id = pybullet.createMultiBody(baseMass=1,
                            baseInertialFramePosition=[0, 0, 0],
                            baseCollisionShapeIndex=collision_id,
                            baseVisualShapeIndex=visual_id,
                            useMaximalCoordinates=False)
    
    pybullet.changeVisualShape(multibody_id, -1, rgbaColor=color)
    return multibody_id, collision_id, visual_id

def any_key_is_pressed():
    keys = pybullet.getKeyboardEvents()
    return any(state & pybullet.KEY_WAS_RELEASED for state in keys.values())

def uniform_size():
    return BIG

def uniform_object():
    return CUBE

def uniform_color():
    return [1, 1, 1, 1]

def make_objects_of_size_and_color(size_func=uniform_size, object_func=uniform_object, color_func=uniform_color):
    current_color = 0
    while (1):
        if any_key_is_pressed():
            spawn_object(object_func, size_func, color_func)

def set_position(body_id, pos):
    pybullet_planning.set_point(body_id, pybullet_planning.Point(pos[0], pos[1], pos[2]))

def create_pybullet_box(pos=[0, 0, 0], width=1, length=1, height=1, color=[1., 1., 1., 1.]):
    # creates target box st. launch_box will launch a box at the target box
    pybullet_planning.create_box(width, length, height, color=color)
    box_geometry = pybullet_planning.get_box_geometry(width, length, height)
    collision_id, visual_id = pybullet_planning.create_shape(geometry=box_geometry)
    body_id = pybullet_planning.create_body(collision_id, visual_id)
    pybullet_planning.set_point(body_id, pybullet_planning.Point(pos[0], pos[1], pos[2]))
    pybullet_planning.set_point(visual_id, pybullet_planning.Point(pos[0], pos[1], pos[2]))
    pybullet_planning.set_point(collision_id, pybullet_planning.Point(pos[0], pos[1], pos[2]))
    pybullet_planning.set_color(body_id, color)
    pybullet_planning.set_color(collision_id, color)
    pybullet_planning.set_color(visual_id, color)
    return body_id

def create_pybullet_cylinder(pos=[0, 0, 0], radius=1, height=1, color=[1., 1., 1., 1.]):
    pybullet_planning.create_cylinder(radius, height, color=color)
    cylinder_geometry = pybullet_planning.get_cylinder_geometry(radius, height)
    collision_id, visual_id = pybullet_planning.create_shape(geometry=cylinder_geometry)
    body_id = pybullet_planning.create_body(collision_id, visual_id)
    pybullet_planning.set_point(body_id, pybullet_planning.Point(pos[0], pos[1], pos[2]))
    pybullet_planning.set_point(visual_id, pybullet_planning.Point(pos[0], pos[1], pos[2]))
    pybullet_planning.set_point(collision_id, pybullet_planning.Point(pos[0], pos[1], pos[2]))
    pybullet_planning.set_color(body_id, color)
    pybullet_planning.set_color(collision_id, color)
    pybullet_planning.set_color(visual_id, color)
    return body_id

def create_pybullet_capsule(pos=[0, 0, 0], radius=1, height=1, color=[1., 1., 1., 1.]):
    pybullet_planning.create_capsule(radius, height, color=color)
    capsule_geometry = pybullet_planning.get_capsule_geometry(radius, height)
    collision_id, visual_id = pybullet_planning.create_shape(geometry=capsule_geometry)
    body_id = pybullet_planning.create_body(collision_id, visual_id)
    pybullet_planning.set_point(body_id, pybullet_planning.Point(pos[0], pos[1], pos[2]))
    pybullet_planning.set_point(visual_id, pybullet_planning.Point(pos[0], pos[1], pos[2]))
    pybullet_planning.set_point(collision_id, pybullet_planning.Point(pos[0], pos[1], pos[2]))
    pybullet_planning.set_color(body_id, color)
    pybullet_planning.set_color(collision_id, color)
    pybullet_planning.set_color(visual_id, color)
    return body_id

def create_pybullet_sphere(pos=[0, 0, 0], radius=1, color=[1., 1., 1., 1.]):
    pybullet_planning.create_sphere(radius, color=color)
    sphere_geometry = pybullet_planning.get_sphere_geometry(radius)
    collision_id, visual_id = pybullet_planning.create_shape(geometry=sphere_geometry)
    body_id = pybullet_planning.create_body(collision_id, visual_id)
    pybullet_planning.set_point(body_id, pybullet_planning.Point(pos[0], pos[1], pos[2]))
    pybullet_planning.set_point(visual_id, pybullet_planning.Point(pos[0], pos[1], pos[2]))
    pybullet_planning.set_point(collision_id, pybullet_planning.Point(pos[0], pos[1], pos[2]))
    pybullet_planning.set_color(body_id, color)
    pybullet_planning.set_color(collision_id, color)
    pybullet_planning.set_color(visual_id, color)
    return body_id


def launch_object():
    DURATION = 10000
    ALPHA = 100000
    # THIS MOVES, BUT NOT WELL
    box_id, _, _ = spawn_object(uniform_object(), uniform_size(), uniform_color())
    cubeStartPos = [0, 0, 1]
    cubeStartOrientation = pybullet.getQuaternionFromEuler([0, 0, 0])

    for i in range(DURATION):
        pybullet.stepSimulation()
        time.sleep(1./240.)
        box_pos, _ = pybullet_planning.get_pose(box_id)
        target_pos = np.random.rand(3) - 0.5
        box_pos = np.array(box_pos)

        force = ALPHA * target_pos
        # to ensure the force is only applied in the xy direction
        force[2] = 0
        pybullet.applyExternalForce(objectUniqueId=box_id, linkIndex=-1, forceObj=force, posObj=box_pos, flags=pybullet.WORLD_FRAME)
    print("Simulation ended normally.")
    pybullet.disconnect()

def set_gravity_z_axis(gravitational_constant):
    pybullet.setGravity(0,0,gravitational_constant)

def set_gravity_all_axes(gravity_for_each_axis):
    pybullet.setGravity(*gravity_for_each_axis)

def set_simulation_time_step(time_step):
    pybullet.setTimeStep(time_step)

def simulation_is_running():
    return pybullet.isConnected()
